
var g_Replays = [];

/**
 * @param {} replaySelectionData 
 * @param {*} compareFiles 
 */
function loadReplays(replaySelectionData, compareFiles)
{
	g_Replays = Engine.GetReplays(compareFiles) || [];

	if (!g_Replays)
		return;

	// g_Playernames = [];
	for (let replay of g_Replays)
	{
		let nonAIPlayers = 0;

		// Check replay for compatibility
		replay.isCompatible = isReplayCompatible(replay);

		sanitizeGameAttributes(replay.attribs);

		// Extract playernames
		for (let playerData of replay.attribs.settings.PlayerData)
		{
			if (!playerData || playerData.AI)
				continue;

			// Remove rating from nick
			// let playername = playerData.Name;
			// let ratingStart = playername.indexOf(" (");
			// if (ratingStart != -1)
			// 	playername = playername.substr(0, ratingStart);

			// if (g_Playernames.indexOf(playername) == -1)
			// 	g_Playernames.push(playername);

			++nonAIPlayers;
        }

		replay.isMultiplayer = nonAIPlayers > 1;

		replay.isRated = nonAIPlayers == 2 &&
			replay.attribs.settings.PlayerData.length == 2 &&
			replay.attribs.settings.RatingEnabled;
	}
}

/**
 * We may encounter malformed replays.
 */
function sanitizeGameAttributes(attribs)
{
	if (!attribs.settings)
		attribs.settings = {};

	if (!attribs.settings.Size)
		attribs.settings.Size = -1;

	if (!attribs.settings.Name)
		attribs.settings.Name = "";

	if (!attribs.settings.PlayerData)
		attribs.settings.PlayerData = [];

	if (!attribs.settings.PopulationCap)
		attribs.settings.PopulationCap = 300;

	if (!attribs.settings.mapType)
		attribs.settings.mapType = "skirmish";

	// Remove gaia
	if (attribs.settings.PlayerData.length && attribs.settings.PlayerData[0] == null)
		attribs.settings.PlayerData.shift();

	attribs.settings.PlayerData.forEach((pData, index) => {
		if (!pData.Name)
			pData.Name = "";
	});
}

/**
 * True if we can start the given replay with the currently loaded mods.
 */
function isReplayCompatible(replay)
{
	return replayHasSameEngineVersion(replay) && hasSameMods(replay.attribs.mods, g_EngineInfo.mods);
}

/**
 * True if we can start the given replay with the currently loaded mods.
 */
function replayHasSameEngineVersion(replay)
{
	return replay.attribs.engine_version && replay.attribs.engine_version == g_EngineInfo.engine_version;
}

function getPlayerData(replay, game, playerRatedData = {})
{	
	if (!replay || !replay.directory)
		return false;	
	let metadata = Engine.GetReplayMetadata(replay.directory);
	if (!metadata)
		return false;

	// 10 minutes gametime minimum
	if (!metadata.timeElapsed || (metadata.timeElapsed / 1000) < 600)
		return false;

	if (!replay.isMultiplayer || !replayEqualsGame(metadata.mapSettings, game))
		return false;

	let players = replay.attribs.settings.PlayerData;

	let states = metadata.playerStates.filter((a, i) => i != 0);

	for (let id in states)
	{
		if (!states[id].sequences || !!players[id].AI)
			continue;

		let state = states[id];
		let seqs = state.sequences;

		// let kills = seqs.enemyUnitsKilledValue;
		// let i = kills.length - 2;
		// for (; i >= 0; --i)
		// 	if (kills[i] != kills[kills.length - 1])
		// 		break;

		// let lastKillIndex = i + 1;

		// let time = seqs.time.length;

		// if (!lastKillIndex) // || lastI < time*0.66)
		// 	continue;

		let lastScoreIndex = seqs.time.length - 1;

		let calculateScoreTotal = (playerState, index) =>
			calculateEconomyScore(playerState, index) + calculateMilitaryScore(playerState, index);

		if (false && splitRatingFromNick(players[id].Name).nick == "aixo")
			for (let i = 0; i < lastScoreIndex; ++i)
				warn(calculateScoreTotal(state, i));

		let score = calculateScoreTotal(state, lastScoreIndex);

		for (; lastScoreIndex >= 0; --lastScoreIndex)
			if (calculateScoreTotal(state, lastScoreIndex) != score)
				break;

		let time = seqs.time.length;
		if (lastScoreIndex == 0 || lastScoreIndex < time*0.66)
			continue;
		
		let scorePerTime = lastScoreIndex ? score/lastScoreIndex : 0;

		if (players[id] && players[id].Name && !players[id].AI)
		{	
			let player = playerRatedData[splitRatingFromNick(players[id].Name).nick] || { "scorePerTime": 0, "games": 0 };
			++player.games;
			player.scorePerTime += scorePerTime;
			playerRatedData[splitRatingFromNick(players[id].Name).nick] = player;
		}
	}

	return true;
}

var gAME = {PlayerData: [null, {Name:"Pudim (1541)", Civ:"brit", Color:{r:50, g:170, b:170}, AI:"", AIDiff:3, AIBehavior:"random", Team:0}, {Name:"Unknown_Player (1588)", Civ:"brit", Color:{r:86, g:180, b:31}, AI:"", AIDiff:3, AIBehavior:"random", Team:0}, {Name:"HMS-Surprise (1361)", Civ:"gaul", Color:{r:21, g:55, b:149}, AI:"", AIDiff:3, AIBehavior:"random", Team:0}, {Name:"mgx (1361)", Civ:"rome", Color:{r:64, g:64, b:64}, AI:"", AIDiff:3, AIBehavior:"random", Team:0}, {Name:"Dakara (1674)", Civ:"gaul", Color:{r:160, g:80, b:200}, AI:"", AIDiff:3, AIBehavior:"random", Team:1}, {Name:"Vasil9555 (1531)", Civ:"gaul", Color:{r:231, g:200, b:5}, AI:"", AIDiff:3, AIBehavior:"random", Team:1}, {Name:"r4pt0r (1334)", Civ:"rome", Color:{r:150, g:20, b:20}, AI:"", AIDiff:3, AIBehavior:"random", Team:1}, {Name:"Havran (1457)", Civ:"iber", Color:{r:220, g:115, b:16}, AI:"", AIDiff:3, AIBehavior:"random", Team:1}], VictoryConditions:["conquest"], CircularMap:true, Size:256, PopulationCap:200, StartingResources:300, Ceasefire:0, RelicCount:2, RelicDuration:20, WonderDuration:20, RegicideGarrison:false, Nomad:false, RevealMap:false, ExploreMap:false, DisableTreasures:false, DisableSpies:false, LockTeams:true, LastManStanding:false, CheatsEnabled:false, Name:"Mainland", Script:"alpine_valley.js", Description:"High Alpine mountains bordering deep valleys.", Keywords:[], Preview:"alpine_valley.png", RatingEnabled:true, Biome: "", TriggerDifficulty: ""};

var g_LookUpGameSettings = 	[ "VictoryConditions", "PopulationCap", "StartingResources", "Ceasefire", "RelicCount", "RelicDuration", "WonderDuration", "RegicideGarrison", "Nomad", "RevealMap", "ExploreMap", "DisableTreasures", "DisableSpies", "LockTeams", "LastManStanding", "CheatsEnabled", "Name", "RatingEnabled" ];

var g_LookUpGameSettingsValid = [ "VictoryConditions", "PopulationCap", "StartingResources", "Nomad",  "DisableTreasures", "LockTeams", "CheatsEnabled" ];

function replayEqualsGame(metadata, game)
{
	let gameTeamsPlayerNum = {};
	let replayTeamsPlayerNum = {};

	let playerGame = game.PlayerData.filter(a => a);
	let playerReplay = metadata.PlayerData.filter(a => a);

	for (let setting of g_LookUpGameSettingsValid)
	{
		if (!setting in gAME || !setting in metadata)
			continue;

		if (!Array.isArray(gAME[setting]))
			if (gAME[setting] != metadata[setting]) {
				// warn(setting + " "+ gAME[setting] + " " + metadata[setting])
				return false;
			}
	}

	if (playerGame.length != playerReplay.length)
		return false;

	playerGame.forEach(a => gameTeamsPlayerNum[a.Team] = (gameTeamsPlayerNum[a.Team] || 0) + 1);
	playerReplay.forEach(a => replayTeamsPlayerNum[a.Team] = (replayTeamsPlayerNum[a.Team] || 0) + 1);
	let gameTeams = Object.keys(gameTeamsPlayerNum);
	let replayTeams = Object.keys(replayTeamsPlayerNum);
	if (gameTeams.length != replayTeams.length)
		return false

	for (let i = 0; i < gameTeams.length; ++i)
		if (gameTeamsPlayerNum[gameTeams[i]] != replayTeamsPlayerNum[replayTeams[i]])
			return false;

	return true;

	for (let setting of g_LookUpGameSettings)
	{
		if (!setting in gAME || !setting in metadata)
			continue;

		if (!Array.isArray(gAME[setting]))
			if (gAME[setting] != metadata[setting]) {
				warn(setting + " "+ gAME[setting] + " " + metadata[setting])
				return false;
			}
	}
	return true;
}

var g_ReplaysPlayersRate = {};
var g_ResourceData = new Resources();
