/**
 * Number of milliseconds to display network warnings.
 */
var g_NetworkWarningTimeout = 3000;

/**
 * Currently displayed network warnings. At most one message per user.
 */
var g_NetworkWarnings = {};

var g_PlayersTimeOutsOrBadConnec = {}
var g_PlayersTimeOut= {};
let g_PlayersBadConnec = [];
var g_IsSession = false;

/**
 * Message-types to be displayed.
 */
var g_NetworkWarningTexts = {

	"server-timeout": (msg, username) =>
		sprintf(translate("Losing connection to server (%(seconds)ss)"), {
			"seconds": Math.ceil(msg.lastReceivedTime / 1000)
		}),

	"client-timeout": (msg, username, usernameRaw) => {
		let user = usernameRaw || username;
		// Object.keys(g_PlayerAssignments).find(pl => { if (g_PlayerAssignments[pl].name == user) {
		// 	warn(g_PlayerAssignments[pl].player)
		// 	warn(g_Players[g_PlayerAssignments[pl].player].state) }});

		let autoKickSeconds = Engine.ConfigDB_GetValue("user", "session.specsautokickseconds");

		let now = Date.now();

		if (g_IsSession && autoKickSeconds >= 0 && g_IsController)
		{
			let userData = g_PlayersTimeOut[user] || { "lastTimeOut": now, "timeInTimeOut": 0, "hit": 0, "kicked": false, "msg": false };
			
			if (userData.kicked)
			{
				userData.lastTimeOut = now;
				userData.kicked = false;
				userData.timeInTimeOut = 0;
			}
			else
			{
				if ((now - userData.lastTimeOut) / 1000 > autoKickSeconds)
				{
					userData.lastTimeOut = now;
					userData.timeInTimeOut += msg.lastReceivedTime;
				}
				else
					userData.timeInTimeOut = msg.lastReceivedTime;
			}

			g_PlayersTimeOut[user] = userData;
			let guid = Object.keys(g_PlayerAssignments).filter(player => 
				g_PlayerAssignments[player].name == user)[0] || -1;
			warn(guid)

			const checkPlayerSpectator = () =>
				Object.keys(g_PlayerAssignments).some(player => 
						g_PlayerAssignments[player].name == user &&
						(g_PlayerAssignments[player].player == -1 || (
							g_Players[g_PlayerAssignments[player].player].state &&
							g_Players[g_PlayerAssignments[player].player].state != "active")) ); // != active: ended player (defeated/won)
			// warn(uneval(Object.keys(g_PlayerAssignments[player])));

			if (!g_PlayerJoining[guid] && checkPlayerSpectator())
			if (true && userData.timeInTimeOut / 1000 > autoKickSeconds)
			{
				Engine.SendNetworkChat("/me Auto kick " + user + " due to connection time out (" + autoKickSeconds + " seconds).");// " " +user + " your connection is timed out, auto kick spectators after " + 
				kickPlayer(user, false);
				userData.kicked = true;
			}
			else if (!userData.msg)
			{
				let msgStart = g_IsSession ? "/msg " + user : "";
				Engine.SendNetworkChat(msgStart + " " +user + " your connection is timed out, auto kick spectators after " + autoKickSeconds + " second(s) is enabled.");
				userData.msg = true;
			}
		}

		return sprintf(translate("%(player)s losing connection (%(seconds)ss)"), {
			"player": username,
			"seconds": Math.ceil(msg.lastReceivedTime / 1000)
		});
	},

	"server-latency": (msg, username) =>
		sprintf(translate("Bad connection to server (%(milliseconds)sms)"), {
			"milliseconds": msg.meanRTT
		}),

	"client-latency": (msg, username, usernameRaw) => {

		let specsautokicktimes = Engine.ConfigDB_GetValue("user", "session.specsautokicktimes");

		if (true && g_IsSession && g_IsController && specsautokicktimes >= 0)
		{
			let user = usernameRaw || username;
			let now = Date.now();

			// let idx = g_PlayersBadConnec.findIndex(pl => pl.name == username);
			
			let userData = g_PlayersBadConnec[user] || { "lastLag": now, "hit": 0, "msg": false };

			if (userData.lastLag - now > specsautokicktimes * 1000)
			{
				userData.lastLag = now;
				userData.hit = Math.floor(msg.meanRTT / 500);
			}
			else
			{
				userData.hit +=	Math.floor(msg.meanRTT / 500);
			}
			g_PlayersBadConnec[user] = userData;
			 
			if (userData.hit < specsautokicktimes)
			{
				if (!userData.msg)
				{
					let msgStart = g_IsSession ? "/msg " + user : "";
					Engine.SendNetworkChat(msgStart + " Warning "+ user+": Auto kick bad connection spectators enabled after reported lag of " + specsautokicktimes + " seconds.");
					userData.msg = true;
				}
			}
			else
			{
				kickPlayer(user, false);
				delete g_PlayersBadConnec[user];
			}
		}
				
		return sprintf(translate("Bad connection to %(player)s (%(milliseconds)sms)"), {
			"player": username,
			"milliseconds": msg.meanRTT
		});
	}
};


function reduceBadPingsCounts()
{
	for (let j in g_PlayersBadConnec) { 
		let el = g_PlayersBadConnec[j];
		if (Date.now() - el.last > 2000) // start reducing when no new bad ping alert after 2 seconds
			--el.count;
		if (el.count < 0)
			g_PlayersBadConnec.splice(j, 1);
	}
}

function reduceBadPingsCountsStart()
{ 
	setTimeout(() => { reduceBadPingsCounts(); reduceBadPingsCountsStart() }, 1000);
}

var g_NetworkCommands = {
	"/kick": argument => kickPlayer(argument, false),
	"/ban": argument => kickPlayer(argument, true),
	"/kickspecs": argument => kickObservers(false),
	"/banspecs": argument => kickObservers(true),
	"/list": argument => addChatMessage({ "type": "clientlist" }),
	"/clear": argument => clearChatMessages(),
	"/showip":  argument =>  { 
			if (g_ServerIP && g_ServerPort)
			{
	// 			let game = g_GameList[gamesBox.selected];
	// 			let ip = game.stunIP ? game.stunIP : game.ip;
	// 			let port = game.stunPort ? game.stunPort : game.port;

	// g_ServerIP = initData.serverIP;
	// g_ServerPort = initData.serverPort;
	// g_UseSTUN = initData.useSTUN;
	// g_HostJID = initData.hostJID;
	// g_ServerName = !!initData.attribs && !!initData.attribs.gameName && initData.attribs.gameName;
	// g_PlayerName = !!initData.playerName ? initData.playerName : "";
				addChatMessage({
					"from": "system",
					"type": "system",
					"text": (g_ServerName ? "IP from \"" + escapeText(g_ServerName) + "\"" : "IP") + ": " + g_ServerIP + ":" + g_ServerPort + " (stun: " + (!!g_UseSTUN ? "true" : "false") + ")"
				});
				return true;
			}
			return false;
	}
};

var g_ValidPorts = { "min": 1, "max": 65535 };

function getValidPort(port)
{
	if (isNaN(+port) || +port < g_ValidPorts.min || +port > g_ValidPorts.max)
		return Engine.GetDefaultPort();

	return +port;
}

/**
 * Must be kept in sync with source/network/NetHost.h
 */
function getDisconnectReason(id, wasConnected)
{
	switch (id)
	{
	case 0: return wasConnected ?
		translateWithContext("network disconnect", "Unknown reason") :
		translate("This is often caused by UDP port 20595 not being forwarded on the host side, by a firewall, or anti-virus software");
	case 1: return translate("The host has ended the game");
	case 2: return translate("Incorrect network protocol version");
	case 3: return translate("Game is loading, please try again later");
	case 4: return translate("Game has already started, no observers allowed."); //\n(Maybe buddies as late join observers allowed use your Real Lobby Name For In-Game)");
	case 5: return translate("You have been kicked");
	case 6: return translate("You have been banned");
	case 7: return translate("Playername in use. If you were disconnected, retry in few seconds");
	case 8: return translate("Server full");
	case 9: return translate("Secure lobby authentication failed. Join via lobby");
	default:
		warn("Unknown disconnect-reason ID received: " + id);
		return sprintf(translate("\\[Invalid value %(id)s]"), { "id": id });
	}
}

/**
 * Show the disconnect reason in a message box.
 *
 * @param {number} reason
 */
function reportDisconnect(reason, wasConnected, hostUser)
{
	// Translation: States the reason why the client disconnected from the server.
	let reasonText = sprintf(translate("Reason: %(reason)s."), { "reason": getDisconnectReason(reason, wasConnected) });
	let askHostObserver = "";
	let btns = null; //[];
	let acts = null;

	if (hostUser && reason == 4 && Engine.HasXmppClient())
	{
		askHostObserver = "\n\n(FGod) Ask in the Lobby Hostuser Player \"" + hostUser + "\" for observing his game?";
		btns = [translate("No"), translate("Yes")];
		acts = [function() { warn("ok2"); }, function() { warn("OK"); Engine.LobbySendMessage(hostUser + " can I go into observing your game? (Enable Late Observer Join in General Options)"); return true; }];
	}
	// warn(uneval(acts))
	messageBox(
		400, 200,
		(wasConnected ?
			translate("Lost connection to the server.") :
			translate("Failed to connect to the server.")
		) + "\n\n" + reasonText + askHostObserver,
		translate("Disconnected"),
		[translate("Ok")]
	);
}

function kickError()
{
	addChatMessage({
		"type": "system",
		"text": translate("Only the host can kick clients!")
	});
}

function kickPlayer(username, ban)
{
	if (g_IsController)
		Engine.KickPlayer(username, ban);
	else
		kickError();
}

function kickObservers(ban)
{
	if (!g_IsController)
	{
		kickError();
		return;
	}

	for (let guid in g_PlayerAssignments)
		if (g_PlayerAssignments[guid].player == -1)
			Engine.KickPlayer(g_PlayerAssignments[guid].name, ban);
}

/**
 * Sort GUIDs of connected users sorted by playerindex, observers last.
 * Requires g_PlayerAssignments.
 */
function sortGUIDsByPlayerID()
{
	return Object.keys(g_PlayerAssignments).sort((guidA, guidB) => {

		let playerIdA = g_PlayerAssignments[guidA].player;
		let playerIdB = g_PlayerAssignments[guidB].player;

		if (playerIdA == -1) return +1;
		if (playerIdB == -1) return -1;

		return playerIdA - playerIdB;
	});
}

/**
 * Get a colorized list of usernames sorted by player slot, observers last.
 * Requires g_PlayerAssignments and colorizePlayernameByGUID.
 *
 * @returns {string}
 */
function getUsernameList()
{
	let usernames = sortGUIDsByPlayerID().map(guid => colorizePlayernameByGUID(guid));

	// Translation: Number of currently connected players/observers and their names
	return sprintf(translate("Users (%(num)s): %(users)s"), {
		"users": usernames.join(translate(", ")),
		"num": usernames.length
	});
}

function pingPauser()
{
	let names = g_PausingClients.map(guid => g_PlayerAssignments[guid].name).filter(name => name != g_PlayerAssignments[Engine.GetPlayerGUID()].name).join(", ");
	if (names)
	{
		let msg = { "name": "readyCall", "text": "Can we continue " + names + "?" };
		spamFilter(msg, msg => Engine.SendNetworkChat(msg.text), 5000, (msg, timeLeft) => addChatMessage({ "type": "system", "from": "system", "text": "Spam "+ Math.ceil(timeLeft/1000)+" <= 5 secs" } ) ); 
	}
}

g_NetworkCommands["/pingpauser"] = pingPauser;

/**
 * Execute a command locally. Requires addChatMessage.
 *
 * @param {string} input
 * @returns {Boolean} whether a command was executed
 */
function executeNetworkCommand(input)
{
	if (input.indexOf("/") != 0)
		return false;

	let command = input.split(" ", 1)[0];
	let argument = input.substr(command.length + 1);

	if (g_NetworkCommands[command])
		g_NetworkCommands[command](argument);

	return !!g_NetworkCommands[command];
}

/**
 * Remember this warning for a few seconds.
 * Overwrite previous warnings for this user.
 *
 * @param msg - GUI message sent by NetServer or NetClient
 */
function addNetworkWarning(msg)
{
	if (!g_NetworkWarningTexts[msg.warntype])
	{
		warn("Unknown network warning type received: " + uneval(msg));
		return;
	}

	if (Engine.ConfigDB_GetValue("user", "overlay.netwarnings") != "true")
		return;

	g_NetworkWarnings[msg.guid || "server"] = {
		"added": Date.now(),
		"msg": msg
	};
}

/**
 * Colorizes and concatenates all network warnings.
 * Returns text and textWidth.
 */
function getNetworkWarnings()
{
	// Remove outdated messages
	for (let guid in g_NetworkWarnings)
		if (Date.now() > g_NetworkWarnings[guid].added + g_NetworkWarningTimeout ||
		    guid != "server" && !g_PlayerAssignments[guid])
			delete g_NetworkWarnings[guid];
	
	// Show local messages first
	let guids = Object.keys(g_NetworkWarnings).sort(guid => guid != "server");

	let font = Engine.GetGUIObjectByName("gameStateNotifications").font;

	let messages = [];
	let maxTextWidth = 0;

	for (let guid of guids)
	{
		let msg = g_NetworkWarnings[guid].msg;
		let username = guid != "server" && g_PlayerAssignments[guid].name;

		// Add formatted text
		messages.push(g_NetworkWarningTexts[msg.warntype](msg, colorizePlayernameByGUID(guid), username));

		// Add width of unformatted text
		let textWidth = Engine.GetTextWidth(font, g_NetworkWarningTexts[msg.warntype](msg, username));
		maxTextWidth = Math.max(textWidth, maxTextWidth);
	}

	return {
		"messages": messages,
		"maxTextWidth": maxTextWidth
	};
}
